
splits = 0
parent_splits = 0
fusions = 0
parent_fusions = 0

class Node(object):
    """
    Base Node object. It should be index node
    Each node stores keys and children.

    Attributes:
        parent
    """

    def __init__(self,parent=None):
        """
        Children nodes are stores in values. Parent nodes simply ast as a medium for traversal
        :type parent: Node
        """
        self.keys: list = []
        self.values: list[Node] = []
        self.parent: Node = parent

    def index(self,key):
        """
        Return the index where the key should be
        :type key: str
        """
        for i, item in enumerate(self.keys):
            if key < item:
                return i
        return len(self.keys)

    def __getitem__(self, item):
        return self.values[self.index(item)]

    def __setitem__(self, key, value):
        i = self.index(key)
        self.keys[i:i] = [key]
        self.values.pop(i)
        self.values[i:i] = value

    def __delitem__(self, key):
        i = self.index(key)
        del self.values[i]
        if i < len(self.keys):
            del self.keys[i]
        else:
            del self.keys[i-1]

    def split(self):
        """
        Splits the node into two and stores them as child nodes.
        extract a pivot from the child to be inserted into the keys of the parent.
        @:return key and two children
        """
        global splits, parent_splits
        splits += 1
        parent_splits += 1

        left = Node(self.parent)
        mid = len(self.keys) // 2

        left.keys = self.keys[:mid]
        left.values = self.values[:mid+1]
        for child in left.values:
            child.parent = left

        key = self.keys[mid]
        self.keys = self.keys[mid+1:]
        self.values = self.values[mid+1:]

        return key, [left,self]

    def fusion(self):
        global fusions,parent_fusions
        fusions += 1
        parent_fusions += 1

        index = self.parent.index(self.keys[0])
        # merge this node with next node
        if index < len(self.parent.keys):
            next_node: Node = self.parent.values[index+1]
            next_node.keys[0:0] = self.keys + [self.parent.keys[index]]
            for child in self.values:
                child.parent = next_node
            next_node.values[0:0] = self.values
        else: # if self is last node, merge with prev
            prev: Node = self.parent.values[-2]
            prev.keys += [self.parent.keys[-1]] + self.keys
            for child in self.values:
                child.parent = prev
            prev.values += self.values

    def borrow_key(self, minimum: int):
        index = self.parent.index(self.keys[0])
        if index < len(self.parent.keys):
            next_node: Node = self.parent.values[index+1]
            if len(next_node.keys) > minimum:
                self.keys += [self.parent.keys[index]]

                borrow_node = next_node.values.pop(0)
                borrow_node.parent = self
                self.values += [borrow_node]
                self.parent.keys[index] = next_node.keys.pop(0)
                return True
        elif index != 0:
            prev: Node = self.parent.values[index - 1]
            if len(prev.keys) > minimum:
                self.keys[0:0] = [self.parent.keys[index-1]]

                borrow_node = prev.values.pop()
                borrow_node.parent = self
                self.values[0:0] = [borrow_node]
                self.parent.keys[index-1] = prev.keys.pop()
                return  True

        return False

class Leaf(Node):
    def __init__(self,parent=None,prev_node=None,next_node=None):
        """
        Create a new leaf in the leaf link
        :type prev_node: Leaf
        :type next_node: Leaf
        """
        super(Leaf,self).__init__(parent)
        self.next: Leaf = next_node
        if next_node is not None:
            next_node.prev = self
        self.prev: Leaf = prev_node
        if prev_node is not  None:
            prev_node.next = self

    def __getitem__(self, item):
        return self.values[self.keys.index(item)]

    def __setitem__(self, key, value):
        i = self.index(key)
        if key not in self.keys:
            self.keys[i:i] = [key]
            self.values[i:i] = [value]
        else:
            self.values[i - 1] = value

    def __delitem__(self, key):
        i = self.keys.index(key)
        del self.keys[i]
        del self.values[i]

    def split(self):
        global splits
        splits += 1

        left = Leaf(self.parent,self.prev,self)
        mid = len(self.keys) // 2

        left.keys = self.keys[:mid]
        left.values = self.values[:mid]

        self.keys: list = self.keys[mid:]
        self.values: list = self.values[mid:]

        # when the leaf node is split, set the parent key to the leftmost key of right chlid node
        return self.keys[0], [left,self]

    def fusion(self):
        global fusions
        fusions += 1

        # if self is left sibling, prepend
        if self.next is not  None and self.next.parent == self.parent:
            self.next.keys[0:0] = self.keys
            self.next.values[0:0] = self.values
        # if self is right sibling, append
        else:
            self.prev.keys += self.keys
            self.prev.values += self.values

        if self.next is not None:
            self.next.prev = self.prev
        if self.prev is not None:
            self.prev.next = self.next

    def borrow_key(self, minimum: int):
        index = self.parent.index(self.keys[0])

        if index < len(self.parent.keys) and len(self.next.keys) > minimum:
            self.keys += [self.next.keys.pop(0)]
            self.values += [self.next.values.pop(0)]
            self.parent.keys[index] = self.next.keys[0]
            return True
        elif index != 0 and len(self.prev.keys) > minimum:
            self.keys[0:0] = [self.prev.keys.pop()]
            self.values[0:0] = [self.prev.values.pop()]
            self.parent.keys[index - 1] = self.keys[0]
            return  True

        return False




from nodes import Node, Leaf,splits,parent_splits,fusions,parent_fusions

class BPlusTree(object):
    """
    B+ tree object

    Nodes will be automatically split into two once they're full. When a split occurs
    a key will float upwards and be inserted into the parent node to act as a pivot.
    [a,b,c,d].split() -> Parent[c]
                         |      |
                        [a,b]   [d]
    Attributes:
        maximum (int): The maximum number of keys a node can hold before triggering split.
    """
    root: Node

    def __init__(self,maximum = 4):
        self.root = Leaf()
        self.maximum: int = maximum if maximum > 2 else 2
        self.minimum: int = self.maximum // 2
        self.depth = 0

    def find_end(self) -> Leaf:
        node = self.root
        while True:
            if type(node) is Leaf:
                return node
            else:
                node = node.values[-1]

    def find(self,key) -> Leaf:
        """
        find the leaf
        Returns:
            Leaf: the leaf which should have the key
        """
        node = self.root
        # traverse until the leaf node is reached
        while type(node) is not Leaf:
            node = node[key]

        return node

    def __getitem__(self, item):
        return self.find(item)[item]

    def query(self,key):
        """
        Returns the value for a given key or None if no such key exists
        """
        leaf = self.find(key)
        return leaf[key] if key in leaf.keys else None

    def range(self,start,end):
        start_node = self.find(start)
        end_node = self.find(end)

        current_node = start_node
        cursor = current_node.index(start)
        while  current_node != end_node.next:
            current_key = current_node.keys[cursor]
            cursor += 1

            if current_key >= end:
                break
            elif cursor >= len(current_node.keys):
                cursor = 0
                current_node = current_node.next

            yield current_key

    def cursor(self,start):
        end  = self.find_end().keys[-1]
        for k in self.range(start,end):
            yield k

    def __setitem__(self, key, value,leaf=None):
        """
        Inserts a key-value pair after traversing to a leaf node.
        If the node is full, splits the leaf into two
        """
        if leaf is None:
            leaf = self.find(key)
        leaf[key] = value
        if len(leaf.keys) > self.maximum:
            # NOTE: * is splat;used here to pass multiple returns of split to insert_index
            self.insert_index(*leaf.split())

    def insert(self,key,value):
        """
        Returns :
            (bool,Leaf): the leaf where the key is inserted. return False if the key already exists
        """
        leaf = self.find(key)
        if key in leaf.keys:
            return False,leaf
        else:
            self.__setitem__(key,value,leaf)
            return True,leaf

    def insert_index(self,key,values: list[Node]):
        """
        For a parent and child node,
        insert the values from the child onto the values of parent
        """
        parent = values[1].parent
        if parent is None:
            values[0].parent = values[1].parent = self.root = Node()
            self.depth += 1
            self.root.keys = [key]
            self.root.values = values
            return

        parent[key] = values
        # if node is full, split it
        # once the leaf node is split, it consists of an internal node and two leaf nodes
        # these need to be re-inserted back to the tree
        if len(parent.keys) > self.maximum:
            self.insert_index(*parent.split())


    def delete(self,key,node: Node = None):
        if node is None:
            node = self.find(key)
        del node[key]

        if len(node.keys) < self.minimum:
            if node == self.root:
                if len(self.root.keys) == 0 and len(self.root.values) > 0:
                    self.root = self.root.values[0]
                    self.root.parent = None
                    self.depth -= 1
                return
            elif not node.borrow_key(self.minimum):
                node.fusion()
                self.delete(key,node.parent)

    def show(self,node=None,file=None,_prefix="",_last=True):
        """Prints keys at each level"""
        if node is None:
            node = self.root
        print(_prefix,"`_ " if _last else "|- ", node.keys,sep="",file=file)
        _prefix += "   " if _last else "|  "

        if type(node) is Node:
            # recursively print the key of child nodes if exists
            for i, child in enumerate(node.values):
                _last = (i == len(node.values) - 1)
                self.show(child,file,_prefix,_last)

    def stats(self):
        return splits,parent_splits,fusions,parent_fusions,self.depth
